
import binascii

if __name__ == "__main__":
    from base import SafeSponge
else:
    from safesponge.base import SafeSponge

class Toy(object):
    KeyWidth = 32
    RateWidth = 16
    CapacityWidth = 32
    parts = 8

    def __init__(self,rounds = 2):
        self.rounds = rounds

    def instance(self):
        return SafeSponge(self.KeyWidth,self.RateWidth,self.CapacityWidth,self.parts,self.rounds,self.MixPart,self.CoreRound)

    def rotr(self,b,shift):
        bi = SafeSponge.bytes_to_int(b)
        bi = 0xFF & ((bi>>shift)|(bi<<(8-shift)))
        return SafeSponge.int_to_bytes(bi,8)

    def bit(self,b,p):
        return (b>>p) & 1

    def shuffle(self,s,m):
        o=0
        for j in range(0,8):
            b = self.bit(SafeSponge.bytes_to_int(s),m[j])
            o |= b<<j
        return SafeSponge.int_to_bytes(o,8)

    def Mix8(self,s,i):
        map0 = [0, 4, 1, 5, 2, 6, 3, 7]
        map1 = [3, 2, 7, 6, 1, 0, 5, 4]
        map2 = [1, 0, 3, 2, 5, 4, 7, 6]
        map3 = [4, 5, 6, 7, 0, 1, 2, 3]
        tmp = i & 3
        shift = tmp*2+1
        if i & 4:
            s = self.shuffle(s,map0)
        else:
            s = self.shuffle(s,map1)
        s = self.rotr(s,shift)
        if i & 8:
            s = self.shuffle(s,map2)
        else:
            s = self.shuffle(s,map3)
        return s

    def MixPart(self,s,i):
        for j in range(0,4):
            tmp = i & 0xF
            base = j
            s[base:base+1] = self.Mix8(s[base:base+1],tmp)
            i = i >> 4
        return s

    def CoreRound(self,c,round):
        s_box = (
            0x63, 0x7C, 0x77, 0x7B, 0xF2, 0x6B, 0x6F, 0xC5, 0x30, 0x01, 0x67, 0x2B, 0xFE, 0xD7, 0xAB, 0x76,
            0xCA, 0x82, 0xC9, 0x7D, 0xFA, 0x59, 0x47, 0xF0, 0xAD, 0xD4, 0xA2, 0xAF, 0x9C, 0xA4, 0x72, 0xC0,
            0xB7, 0xFD, 0x93, 0x26, 0x36, 0x3F, 0xF7, 0xCC, 0x34, 0xA5, 0xE5, 0xF1, 0x71, 0xD8, 0x31, 0x15,
            0x04, 0xC7, 0x23, 0xC3, 0x18, 0x96, 0x05, 0x9A, 0x07, 0x12, 0x80, 0xE2, 0xEB, 0x27, 0xB2, 0x75,
            0x09, 0x83, 0x2C, 0x1A, 0x1B, 0x6E, 0x5A, 0xA0, 0x52, 0x3B, 0xD6, 0xB3, 0x29, 0xE3, 0x2F, 0x84,
            0x53, 0xD1, 0x00, 0xED, 0x20, 0xFC, 0xB1, 0x5B, 0x6A, 0xCB, 0xBE, 0x39, 0x4A, 0x4C, 0x58, 0xCF,
            0xD0, 0xEF, 0xAA, 0xFB, 0x43, 0x4D, 0x33, 0x85, 0x45, 0xF9, 0x02, 0x7F, 0x50, 0x3C, 0x9F, 0xA8,
            0x51, 0xA3, 0x40, 0x8F, 0x92, 0x9D, 0x38, 0xF5, 0xBC, 0xB6, 0xDA, 0x21, 0x10, 0xFF, 0xF3, 0xD2,
            0xCD, 0x0C, 0x13, 0xEC, 0x5F, 0x97, 0x44, 0x17, 0xC4, 0xA7, 0x7E, 0x3D, 0x64, 0x5D, 0x19, 0x73,
            0x60, 0x81, 0x4F, 0xDC, 0x22, 0x2A, 0x90, 0x88, 0x46, 0xEE, 0xB8, 0x14, 0xDE, 0x5E, 0x0B, 0xDB,
            0xE0, 0x32, 0x3A, 0x0A, 0x49, 0x06, 0x24, 0x5C, 0xC2, 0xD3, 0xAC, 0x62, 0x91, 0x95, 0xE4, 0x79,
            0xE7, 0xC8, 0x37, 0x6D, 0x8D, 0xD5, 0x4E, 0xA9, 0x6C, 0x56, 0xF4, 0xEA, 0x65, 0x7A, 0xAE, 0x08,
            0xBA, 0x78, 0x25, 0x2E, 0x1C, 0xA6, 0xB4, 0xC6, 0xE8, 0xDD, 0x74, 0x1F, 0x4B, 0xBD, 0x8B, 0x8A,
            0x70, 0x3E, 0xB5, 0x66, 0x48, 0x03, 0xF6, 0x0E, 0x61, 0x35, 0x57, 0xB9, 0x86, 0xC1, 0x1D, 0x9E,
            0xE1, 0xF8, 0x98, 0x11, 0x69, 0xD9, 0x8E, 0x94, 0x9B, 0x1E, 0x87, 0xE9, 0xCE, 0x55, 0x28, 0xDF,
            0x8C, 0xA1, 0x89, 0x0D, 0xBF, 0xE6, 0x42, 0x68, 0x41, 0x99, 0x2D, 0x0F, 0xB0, 0x54, 0xBB, 0x16,
        )
        c[0] ^= round
        for j in range(0,4):
            c[j] = s_box[c[j]]
        t = SafeSponge.int_to_bytes(0,32);
        for j in range(0,4):
            t[j]=0
            for i in range(0,8):
                ci = i%4
                cs = j*2+(i//4)
                assert(cs<8)
                t[j] = (t[j] << 1) | ((c[ci]>>cs) & 1)
        for j in range(0,4):
            c[j] ^= s_box[t[j]]
        return c

    def CoreRound_strong(self,c,round):
        import hashlib
        c = hashlib.sha256(c).digest()[0:4]
        return c

if __name__ == "__main__":
    impl = Toy().instance()
    def test_hash_mode():
        import collision_test
        impl.Verbose(SafeSponge.SPY_F_IO)

        impl.hash(binascii.unhexlify(b''))
        impl.hash(binascii.unhexlify(b'00'))
        impl.hash(binascii.unhexlify(b'0001'))
        impl.hash(binascii.unhexlify(b'0000'))
        impl.hash(binascii.unhexlify(b'000000'))
        impl.Verbose(SafeSponge.SPY_ALG_IO)

        impl.hash(binascii.unhexlify(b'01'))
        impl.hash(binascii.unhexlify(b'02'))
        impl.hash(binascii.unhexlify(b'03'))
        collision_test.collision_test(impl,24,10)

    def aead_enc_dec(key,nonce,message=None,ad=None):
        if message is None:
            message = bytearray(binascii.unhexlify(b''))
        c = impl.encrypt(key,nonce,message,ad)
        p = impl.decrypt(key,nonce,c,ad)
        assert(p==message)

    impl.Verbose(SafeSponge.SPY_ALG_IO)
    key = binascii.unhexlify(b'00')*impl.key_size()
    nonce = binascii.unhexlify(b'00')*impl.block_size()
    aead_enc_dec(key,nonce)
    aead_enc_dec(key,nonce,binascii.unhexlify(b'00'))
    aead_enc_dec(key,nonce,binascii.unhexlify(b'0001'))
    aead_enc_dec(key,nonce,binascii.unhexlify(b'0000'))
    aead_enc_dec(key,nonce,binascii.unhexlify(b'000000'))
    aead_enc_dec(key,nonce,binascii.unhexlify(b'00000001'))
